import os
import time
import traceback
import uuid

import actionlib
import geometry_msgs
import numpy as np
import rospy
from graspit_interface.msg import (PlanGraspsAction, PlanGraspsGoal, Planner, SearchContact, SearchSpace, SimAnnParams)
from graspit_interface.srv import (ApproachToContact, AutoGrasp, AutoOpen, ClearWorld, ComputeEnergy, ComputeQuality,
                                   DynamicAutoGraspComplete, FindInitialContact, ForceRobotDOF, GetBodies, GetBody,
                                   GetDynamics, GetGraspableBodies, GetGraspableBody, GetImage, GetRobot, GetRobots,
                                   ImportGraspableBody, ImportObstacle, ImportRobot, LoadWorld, MoveDOFToContacts,
                                   SaveImage, SaveWorld, SetBodyPose, SetDynamics, SetGraspableBodyPose,
                                   SetRobotDesiredDOF, SetRobotPose, ToggleAllCollisions)
from rospy.exceptions import ROSException

from config import d
from graspit_exceptions import (BodyCollisionException, ClearWorldException, GraspableBodyCollisionException,
                                GraspitTimeoutException, ImportGraspableBodyException, ImportObstacleException,
                                ImportRobotException, InvalidBodyIDException, InvalidBodyPoseException,
                                InvalidDynamicsModeException, InvalidEnergyTypeException,
                                InvalidGraspableBodyIDException, InvalidGraspableBodyPoseException,
                                InvalidRobotDOFCountMismatchException, InvalidRobotDOFOutOfRangeException,
                                InvalidRobotIDException, InvalidRobotPoseException, LoadWorldException,
                                RobotCollisionException, SaveImageException, SaveWorldException)


def _wait_for_service(serviceName, timeout=1, error=False):
    done = False
    while not done:
        try:
            rospy.wait_for_service(serviceName, timeout=timeout)
            done = True
        except ROSException, e:
            if error:
                raise NotImplementedError
            traceback.print_exc()
            time.sleep(1)


def _wait_for_action_server(client, timeout=1):
    try:
        client.wait_for_server(timeout=rospy.Duration(timeout))
    except ROSException, e:
        raise GraspitTimeoutException()


def convertQImageToMat(incomingImage):
    '''  Converts a QImage into an opencv MAT format  '''

    incomingImage = incomingImage.convertToFormat(4)

    width = incomingImage.width()
    height = incomingImage.height()

    ptr = incomingImage.bits()
    ptr.setsize(incomingImage.byteCount())
    arr = np.array(ptr).reshape(height, width, 4)  # Copies the data
    return arr


class GraspitCommander(object):
    """
    Python interface for interacting with GraspIt
    """

    def __init__(self, host, port):
        import os
        self.ROS_NODE_NAME = "GraspItCommanderNode"
        self.GRASPIT_NODE_NAME = "/graspit/"
        os.environ['ROS_MASTER_URI'] = 'http://%s:%s' % (host, port)

    def getRobots(self):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getRobots')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getRobots', GetRobots)
        robotList = serviceProxy()

        return robotList

    def getRobot(self, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getRobot')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getRobot', GetRobot)
        robot = serviceProxy(id)

        if robot.result is GetRobot._response_class.RESULT_SUCCESS:
            return robot
        elif robot.result is GetRobot._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)

    def setRobotPose(self, pose, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'setRobotPose')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'setRobotPose', SetRobotPose)
        result = serviceProxy(id, pose)

        if result.result is SetRobotPose._response_class.RESULT_SUCCESS:
            return
        elif result.result is SetRobotPose._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)
        elif result.result is SetRobotPose._response_class.RESULT_INVALID_POSE:
            raise InvalidRobotPoseException()
        elif result.result is SetRobotPose._response_class.RESULT_ROBOT_IN_COLLISION:
            raise RobotCollisionException()

    def getGraspableBodies(self):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getGraspableBodies')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getGraspableBodies', GetGraspableBodies)
        graspableBodyList = serviceProxy()

        return graspableBodyList

    def getGraspableBody(self, id):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getGraspableBody')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getGraspableBody', GetGraspableBody)
        graspableBody = serviceProxy(id)

        if graspableBody.result is GetGraspableBody._response_class.RESULT_SUCCESS:
            return graspableBody
        elif graspableBody.result is GetGraspableBody._response_class.RESULT_INVALID_ID:
            raise InvalidGraspableBodyIDException(id)

    def setGraspableBodyPose(self, id, pose):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'setGraspableBodyPose')
        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'setGraspableBodyPose', SetGraspableBodyPose)
        result = serviceProxy(id, pose)
        if result.result is SetGraspableBodyPose._response_class.RESULT_SUCCESS:
            return
        elif result.result is SetGraspableBodyPose._response_class.RESULT_INVALID_ID:
            raise InvalidGraspableBodyIDException(id)
        elif result.result is SetGraspableBodyPose._response_class.RESULT_INVALID_POSE:
            raise InvalidGraspableBodyPoseException()
        elif result.result is SetGraspableBodyPose._response_class.RESULT_BODY_IN_COLLISION:
            raise GraspableBodyCollisionException()

    def getBodies(self):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getBodies')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getBodies', GetBodies)
        bodyList = serviceProxy()

        return bodyList

    def getBody(self, id):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getBody')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getBody', GetBody)
        body = serviceProxy(id)

        if body.result is GetBody._response_class.RESULT_SUCCESS:
            return body
        elif body.result is GetBody._response_class.RESULT_INVALID_ID:
            raise InvalidBodyIDException(id)

    def setBodyPose(self, id, pose):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'setBodyPose')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'setBodyPose', SetBodyPose)
        result = serviceProxy(id, pose)

        if result.result is SetBodyPose._response_class.RESULT_SUCCESS:
            return
        elif result.result is SetBodyPose._response_class.RESULT_INVALID_ID:
            raise InvalidBodyIDException(id)
        elif result.result is SetBodyPose._response_class.RESULT_INVALID_POSE:
            raise InvalidBodyPoseException()
        elif result.result is SetBodyPose._response_class.RESULT_ROBOT_IN_COLLISION:
            raise BodyCollisionException()

    def getDynamics(self):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getDynamics')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getDynamics', GetDynamics)
        result = serviceProxy()

        return result.dynamicsEnabled

    def setDynamics(self, enabled):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'setDynamics')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'setDynamics', SetDynamics)
        serviceProxy(enabled)

    def autoGrasp(self, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'autoGrasp')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'autoGrasp', AutoGrasp)
        result = serviceProxy(id)

        if result.result is AutoGrasp._response_class.RESULT_SUCCESS:
            return
        elif result.result is AutoGrasp._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)

    def autoOpen(self, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'autoOpen')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'autoOpen', AutoOpen)
        result = serviceProxy(id)

        if result.result is AutoOpen._response_class.RESULT_SUCCESS:
            return
        elif result.result is AutoOpen._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)

    def forceRobotDof(self, dofs, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'forceRobotDof')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'forceRobotDof', ForceRobotDOF)
        result = serviceProxy(id, dofs)

        if result.result is ForceRobotDOF._response_class.RESULT_SUCCESS:
            return
        elif result.result is ForceRobotDOF._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)
        elif result.result is ForceRobotDOF._response_class.RESULT_DOF_OUT_OF_RANGE:
            raise InvalidRobotDOFOutOfRangeException()
        elif result.result is ForceRobotDOF._response_class.RESULT_DOF_COUNT_MISMATCH:
            raise InvalidRobotDOFCountMismatchException()
        elif result.result is ForceRobotDOF._response_class.RESULT_DYNAMICS_MODE_ENABLED:
            raise InvalidDynamicsModeException()

    def moveDOFToContacts(self, dofs, desired_steps, stop_at_contact, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'moveDOFToContacts')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'moveDOFToContacts', MoveDOFToContacts)
        result = serviceProxy(id, dofs, desired_steps, stop_at_contact)

        if result.result is MoveDOFToContacts._response_class.RESULT_SUCCESS:
            return
        elif result.result is MoveDOFToContacts._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)
        elif result.result is MoveDOFToContacts._response_class.RESULT_DOF_OUT_OF_RANGE:
            raise InvalidRobotDOFOutOfRangeException()
        elif result.result is MoveDOFToContacts._response_class.RESULT_DOF_COUNT_MISMATCH:
            raise InvalidRobotDOFCountMismatchException()
        elif result.result is MoveDOFToContacts._response_class.RESULT_DYNAMICS_MODE_ENABLED:
            raise InvalidDynamicsModeException()

    def setRobotDesiredDOF(self, dofs, dof_velocities, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'setRobotDesiredDOF')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'setRobotDesiredDOF', SetRobotDesiredDOF)
        result = serviceProxy(id, dofs, dof_velocities)

        if result.result is SetRobotDesiredDOF._response_class.RESULT_SUCCESS:
            return
        elif result.result is SetRobotDesiredDOF._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)
        elif result.result is SetRobotDesiredDOF._response_class.RESULT_DOF_OUT_OF_RANGE:
            raise InvalidRobotDOFOutOfRangeException()
        elif result.result is SetRobotDesiredDOF._response_class.RESULT_DOF_COUNT_MISMATCH:
            raise InvalidRobotDOFCountMismatchException()

    def clearWorld(self, error=False):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'clearWorld', error=error)

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'clearWorld', ClearWorld)
        result = serviceProxy()

        if result.result is ClearWorld._response_class.RESULT_SUCCESS:
            return
        elif result.result is ClearWorld._response_class.RESULT_FAILURE:
            raise ClearWorldException()

    def loadWorld(self, worldFile):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'loadWorld')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'loadWorld', LoadWorld)
        result = serviceProxy(worldFile)

        if result.result is LoadWorld._response_class.RESULT_SUCCESS:
            return
        elif result.result is LoadWorld._response_class.RESULT_FAILURE:
            raise LoadWorldException()

    def planGrasps(self, cls, graspable_body_id=0, planner=Planner(Planner.SIM_ANN),
                   search_energy="GUIDED_POTENTIAL_QUALITY_ENERGY",
                   search_space=SearchSpace(SearchSpace.SPACE_AXIS_ANGLE),
                   search_contact=SearchContact(SearchContact.CONTACT_PRESET), max_steps=70000, feedback_cb=None,
                   feedback_num_steps=-1, sim_ann_params=SimAnnParams()):
        try:
            rospy.init_node(cls.ROS_NODE_NAME, anonymous=True)
        except ROSException:
            pass

        client = actionlib.SimpleActionClient(self.GRASPIT_NODE_NAME + 'planGrasps', PlanGraspsAction)
        _wait_for_action_server(client)

        goal = PlanGraspsGoal(graspable_body_id=graspable_body_id, planner=planner, search_energy=search_energy,
                              search_space=search_space, search_contact=search_contact, max_steps=max_steps,
                              feedback_num_steps=feedback_num_steps, sim_ann_params=sim_ann_params)

        client.send_goal(goal, feedback_cb=feedback_cb)
        client.wait_for_result()

        return client.get_result()

    def approachToContact(self, moveDist=200, oneStep=False, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'approachToContact')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'approachToContact', ApproachToContact)
        result = serviceProxy(moveDist, oneStep, id)

        if result.result is ApproachToContact._response_class.RESULT_SUCCESS:
            return
        elif result.result is ApproachToContact._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)

    def findInitialContact(self, id=0, moveDist=200):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'findInitialContact')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'findInitialContact', FindInitialContact)
        result = serviceProxy(id, moveDist)

        if result.result is FindInitialContact._response_class.RESULT_SUCCESS:
            return
        elif result.result is FindInitialContact._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)
        elif result.result is FindInitialContact._response_class.RESULT_BODY_IN_COLLISION:
            raise InvalidRobotPoseException()

    def computeQuality(self, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'computeQuality')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'computeQuality', ComputeQuality)
        result = serviceProxy(id)

        if result.result is ComputeQuality._response_class.RESULT_SUCCESS:
            return result
        elif result.result is ComputeQuality._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)
        elif result.result is ComputeQuality._response_class.RESULT_COLLISION:
            raise InvalidRobotPoseException()

    def computeEnergy(self, energy_type, hand_id=0, body_id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'computeEnergy')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'computeEnergy', ComputeEnergy)
        result = serviceProxy(hand_id, body_id, energy_type)

        if result.result is ComputeEnergy._response_class.RESULT_SUCCESS:
            return result
        elif result.result is ComputeEnergy._response_class.RESULT_INVALID_HAND_ID:
            raise InvalidRobotIDException(hand_id)
        elif result.result is ComputeEnergy._response_class.RESULT_INVALID_BODY_ID:
            raise InvalidGraspableBodyIDException(body_id)
        elif result.result is ComputeEnergy._response_class.RESULT_INVALID_ENERGY_TYPE:
            raise InvalidEnergyTypeException(energy_type)

    def dynamicAutoGraspComplete(self, id=0):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'dynamicAutoGraspComplete')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'dynamicAutoGraspComplete', DynamicAutoGraspComplete)
        result = serviceProxy(id)

        if result.result is DynamicAutoGraspComplete._response_class.RESULT_SUCCESS:
            return result.GraspComplete
        elif result.result is DynamicAutoGraspComplete._response_class.RESULT_INVALID_ID:
            raise InvalidRobotIDException(id)

    def importObstacle(self, obstacleName, pose=None):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'importObstacle')

        if not pose:
            pose = geometry_msgs.msg.Pose()
            pose.orientation.w = 1

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'importObstacle', ImportObstacle)
        result = serviceProxy(obstacleName, pose)

        if result.result is ImportObstacle._response_class.RESULT_SUCCESS:
            return
        elif result.result is ImportObstacle._response_class.RESULT_FAILURE:
            raise ImportObstacleException(name=obstacleName)

    def importGraspableBody(self, bodyName, pose=None):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'importGraspableBody')

        if not pose:
            pose = geometry_msgs.msg.Pose()
            pose.orientation.w = 1

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'importGraspableBody', ImportGraspableBody)
        result = serviceProxy(bodyName, pose)
        if result.result is ImportGraspableBody._response_class.RESULT_SUCCESS:
            return
        elif result.result is ImportGraspableBody._response_class.RESULT_FAILURE:
            raise ImportGraspableBodyException(name=bodyName)

    def importRobot(self, robotName, pose=None):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'importRobot')

        if not pose:
            pose = geometry_msgs.msg.Pose()
            pose.orientation.w = 1

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'importRobot', ImportRobot)
        result = serviceProxy(robotName, pose)

        if result.result is ImportRobot._response_class.RESULT_SUCCESS:
            return 0
        elif result.result is ImportRobot._response_class.RESULT_FAILURE:
            raise ImportRobotException(name=robotName)

    def saveImage(self, fileName):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'saveImage')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'saveImage', SaveImage)
        result = serviceProxy(fileName[:-4], 0, 0, d.depth)

        if result.result is SaveImage._response_class.RESULT_SUCCESS:
            return
        elif result.result is SaveImage._response_class.RESULT_FAILURE:
            raise SaveImageException()

    def getImage(self, shape):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'getImage')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'getImage', GetImage)
        result = serviceProxy(shape[0], shape[1])
        r = np.array(result.r).reshape(shape)
        g = np.array(result.g).reshape(shape)
        b = np.array(result.b).reshape(shape)
        image = np.stack((r, g, b), axis=2)
        return image

    def getImageDisk(self, shape, host, port):
        time.sleep(5)
        _wait_for_service(self.GRASPIT_NODE_NAME + 'saveImage')

        hash = str(uuid.uuid4())
        save_depth_dir = os.path.join(d.server_graspit_img_dir, 'tmp_depth_%s_%s_%s.npy' % (host, port, hash))

        load_depth_dir = os.path.join(os.path.expanduser('~/Desktop/images'), host,
                                      os.path.basename(save_depth_dir)) if host != 'localhost' else save_depth_dir
        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'saveImage', SaveImage)
        result = serviceProxy(save_depth_dir, shape[0], shape[1], d.depth, d.depth_scale, d.depth_bias)

        if result.result is SaveImage._response_class.RESULT_SUCCESS:
            return np.load(load_depth_dir)[:, :, np.newaxis]
        elif result.result is SaveImage._response_class.RESULT_FAILURE:
            raise SaveImageException()

    def saveWorld(self, fileName):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'saveWorld')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'saveWorld', SaveWorld)
        result = serviceProxy(fileName)

        if result.result is SaveWorld._response_class.RESULT_SUCCESS:
            return
        elif result.result is SaveWorld._response_class.RESULT_FAILURE:
            raise SaveWorldException()

    def toggleAllCollisions(self, enableCollisions):
        _wait_for_service(self.GRASPIT_NODE_NAME + 'toggleAllCollisions')

        serviceProxy = rospy.ServiceProxy(self.GRASPIT_NODE_NAME + 'toggleAllCollisions', ToggleAllCollisions)
        serviceProxy(enableCollisions)
